import { AbstractControl, VALID } from './abstract-control';
function find(control, path, delimiter) {
    if (path == null)
        return null;
    if (!Array.isArray(path)) {
        path = path.split(delimiter);
    }
    if (Array.isArray(path) && path.length === 0)
        return null;
    // Not using Array.reduce here due to a Chrome 80 bug
    // https://bugs.chromium.org/p/chromium/issues/detail?id=1049982
    let controlToFind = control;
    path.forEach((name) => {
        if (controlToFind instanceof FormGroup) {
            controlToFind = name in controlToFind.controls ? controlToFind.controls[name] : null;
        }
        else {
            controlToFind = null;
        }
    });
    return controlToFind;
}
export class FormGroup extends AbstractControl {
    constructor(controls, validators, asyncValidator) {
        super(validators !== null && validators !== void 0 ? validators : null, asyncValidator !== null && asyncValidator !== void 0 ? asyncValidator : null);
        Object.defineProperty(this, "controls", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: controls
        });
        Object.defineProperty(this, "_value", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_status", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: VALID
        });
        this._setUpControls();
        this.updateValueAndValidity(true);
    }
    get(path) {
        return find(this, path, '.');
    }
    getError(errorCode, path) {
        const control = path ? this.get(path) : this;
        return control && control.errors ? control.errors[errorCode] : null;
    }
    hasError(errorCode, path) {
        return !!this.getError(errorCode, path);
    }
    addControl(name, control) {
        if (!(name in this.controls)) {
            this.controls[name] = control;
            control.setParent(this);
        }
        this.updateValueAndValidity();
    }
    removeControl(name) {
        delete this.controls[name];
        this.updateValueAndValidity();
    }
    setControl(name, control) {
        delete this.controls[name];
        this.addControl(name, control);
        this.updateValueAndValidity();
    }
    contains(controlName) {
        return controlName in this.controls;
    }
    setValue(value, onlySelf) {
        this._checkAllValuesPresent(value);
        Object.keys(value).forEach((name) => {
            this._throwIfControlMissing(name);
            this.controls[name].setValue(value[name], true);
        });
        this.updateValueAndValidity(onlySelf);
    }
    patchValue(value, onlySelf) {
        Object.keys(value).forEach((name) => {
            if (this.controls[name]) {
                this.controls[name].patchValue(value[name], true);
            }
        });
        this.updateValueAndValidity(onlySelf);
    }
    reset(value = {}, onlySelf) {
        this._forEachChild((control, name) => {
            control.reset(value[name], true);
        });
        this._updatePristine(onlySelf);
        this.updateValueAndValidity(onlySelf);
    }
    _updateValue() {
        this._value = this._reduceValue();
    }
    _forEachChild(cb) {
        Object.keys(this.controls).forEach((key) => {
            // The list of controls can change (for ex. controls might be removed) while the loop
            // is running (as a result of invoking Forms API in `valueChanges` subscription), so we
            // have to null check before invoking the callback.
            const control = this.controls[key];
            control && cb(control, key);
        });
    }
    _anyControls(condition) {
        for (const controlName of Object.keys(this.controls)) {
            const control = this.controls[controlName];
            if (this.contains(controlName) && condition(control)) {
                return true;
            }
        }
        return false;
    }
    _allControlsDisabled() {
        for (const controlName of Object.keys(this.controls)) {
            if (this.controls[controlName].enabled) {
                return false;
            }
        }
        return Object.keys(this.controls).length > 0 || this.disabled;
    }
    _setUpControls() {
        this._forEachChild((control) => {
            control.setParent(this);
        });
    }
    _reduceValue() {
        return this._reduceChildren({}, (acc, control, name) => {
            acc[name] = control.value;
            return acc;
        });
    }
    _reduceChildren(initValue, fn) {
        let res = initValue;
        this._forEachChild((control, name) => {
            res = fn(res, control, name);
        });
        return res;
    }
    _checkAllValuesPresent(value) {
        this._forEachChild((control, name) => {
            if (value[name] === undefined) {
                throw new Error(`Must supply a value for form control with name: '${name}'.`);
            }
        });
    }
    _throwIfControlMissing(name) {
        if (!Object.keys(this.controls).length) {
            throw new Error(`
          There are no form controls registered with this group yet. If you're using ngModel,
          you may want to check next tick (e.g. use setTimeout).
        `);
        }
        if (!this.controls[name]) {
            throw new Error(`Cannot find form control with name: ${name}.`);
        }
    }
}
